# -*- coding: utf-8 -*-

from omi.dialects.oep.parser import JSONParser_1_3, JSONParser_1_4
from ..internal_structures import metadata_v_1_3, metadata_v_1_4
from ..base.parser import _test_generic_parsing


def test_parser_v1_3():
    parser = JSONParser_1_3()
    _input_file = "tests/data/metadata_v13.json"
    expected_result = metadata_v_1_3
    _test_generic_parsing(parser, _input_file, expected_result, nulls=["none", ""])


def test_parser_v1_4():
    parser = JSONParser_1_4()
    _input_file = "tests/data/metadata_v14.json"
    expected_result = metadata_v_1_4
    _test_generic_parsing(parser, _input_file, expected_result)





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json

from dateutil.parser import parse as parse_date

from omi import structure
from omi.dialects.base.parser import Parser


class JSONParser(Parser):
    def load_string(self, string: str, *args, **kwargs):
        return json.loads(string)

    def is_valid(self, inp: str):
        """Checks the validity of a JSON string

        Parameters
        ----------
        inp: str
            The JSON string to be checked.

        Returns
        -------
        bool
            True if valid JSON, False otherwise.
        """

        try:
            json.loads(inp)
        except ValueError:
            return False
        return True


class JSONParser_1_3(JSONParser):
    def is_valid(self, inp: str):
        if not super(self, JSONParser_1_3).is_valid(inp):
            return False
        try:
            self.assert_1_3_metastring(inp)
        except:
            return False
        else:
            return True

    def parse(self, json_old, *args, **kwargs):
        # context section
        context = structure.Context(
            homepage=None,
            documentation=None,
            source_code=None,
            contact=None,
            grant_number=None,
        )

        # filling the spatial section
        old_spatial = json_old.get("spatial")
        spatial = structure.Spatial(
            location=None,
            extent=old_spatial.get("extent"),
            resolution=old_spatial.get("resolution"),
        )

        # filling the temporal section
        temporal = structure.Temporal(
            reference_date=parse_date(json_old["temporal"].get("reference_date")),
            start=None,
            end=None,
            resolution=None,
            ts_orientation=None,
        )

        # filling the source section
        sources = [
            structure.Source(
                title=old_source.get("name"),
                description=old_source.get("description"),
                path=old_source.get("url"),
                source_license=None,
                source_copyright=old_source.get("copyright"),
            )
            for old_source in json_old.get("sources")
        ]

        # filling the license section
        old_license = json_old.get("license")
        licenses = [
            structure.TermsOfUse(
                lic=structure.License(
                    identifier=old_license.get("id"),
                    name=old_license.get("name"),
                    path=old_license.get("url"),
                    other_references=[],
                    text=None,
                ),
                instruction=old_license.get("instruction"),
                attribution=old_license.get("copyright"),
            )
        ]

        # filling the contributers section
        contributions = [
            structure.Contribution(
                contributor=structure.Person(
                    name=old_contributor.get("name"), email=old_contributor.get("email")
                ),
                date=parse_date(old_contributor.get("date")),
                obj=None,
                comment=old_contributor.get("comment"),
            )
            for old_contributor in json_old.get("contributors")
        ]

        # extending with script-user information

        resources = []
        for resource in json_old.get("resources"):
            fields = [
                structure.Field(
                    name=field.get("name"),
                    description=field.get("description"),
                    field_type=None,
                    unit=field.get("unit"),
                )
                for field in resource.get("fields", [])
            ]
            schema = structure.Schema(fields=fields, primary_key=None, foreign_keys=[])
            resources.append(
                structure.Resource(
                    profile=None,
                    name=resource.get("name"),
                    path=None,
                    resource_format="PostgreSQL",
                    encoding=None,
                    dialect=None,
                    schema=schema,
                )
            )

        review = structure.Review(path=None, badge=None)

        comment = structure.MetaComment(
            metadata_info="Metadata documentation and explanation (https://github.com/OpenEnergyPlatform/organisation/wiki/metadata)",
            dates="Dates and time must follow the ISO8601 including time zone (YYYY-MM-DD or YYYY-MM-DDThh:mm:ss±hh)",
            units="Use a space between numbers and units (100 m)",
            languages="Languages must follow the IETF (BCP47) format (en-GB, en-US, de-DE)",
            licenses="License name must follow the SPDX License List (https://spdx.org/licenses/",
            review="Following the OEP Data Review (https://github.com/OpenEnergyPlatform/data-preprocessing/wiki)",
            none="If not applicable use (none)",
        )

        metadata = structure.OEPMetadata(
            name=None,
            title=json_old.get("title"),
            identifier=None,
            description=json_old.get("description"),
            languages=json_old.get("language"),
            keywords=[],
            publication_date=None,
            context=context,
            spatial=spatial,
            temporal=temporal,
            sources=sources,
            terms_of_use=licenses,
            contributions=contributions,
            resources=resources,
            review=review,
            comment=comment,
        )
        return metadata


class JSONParser_1_4(JSONParser):
    def is_valid(self, inp: str):
        if not super(self, JSONParser_1_4).is_valid(inp):
            return False
        try:
            self.assert_1_3_metastring(inp)
        except:
            return False
        else:
            return True

    def parse(self, json_old):
        # context section
        inp_context = json_old.get("context")
        context = structure.Context(
            homepage=inp_context.get("homepage"),
            documentation=inp_context.get("documentation"),
            source_code=inp_context.get("sourceCode"),
            contact=inp_context.get("contact"),
            grant_number=inp_context.get("grantNo"),
        )

        # filling the spatial section
        old_spatial = json_old.get("spatial")
        spatial = structure.Spatial(
            location=old_spatial.get("location"),
            extent=old_spatial.get("extent"),
            resolution=old_spatial.get("resolution"),
        )

        # filling the temporal section
        inp_temporal = json_old["temporal"]
        temporal = structure.Temporal(
            reference_date=parse_date(inp_temporal.get("referenceDate")),
            start=parse_date(inp_temporal.get("start")),
            end=parse_date(inp_temporal.get("end")),
            resolution=inp_temporal.get("resolution"),
            ts_orientation=structure.TimestampOrientation.create(
                inp_temporal.get("timestamp")
            ),
        )

        # filling the source section
        sources = [
            structure.Source(
                title=old_source.get("title"),
                description=old_source.get("description"),
                path=old_source.get("path"),
                source_license=structure.License(
                    name=None,
                    identifier=old_source.get("license"),
                    other_references=[],
                    path=None,
                    text=None,
                ),
                source_copyright=old_source.get("copyright"),
            )
            for old_source in json_old.get("sources", [])
        ]

        # filling the license section
        licenses = [
            structure.TermsOfUse(
                lic=structure.License(
                    identifier=old_license.get("name"),
                    name=old_license.get("title"),
                    path=old_license.get("path"),
                    other_references=[],
                    text=None,
                ),
                instruction=old_license.get("instruction"),
                attribution=old_license.get("attribution"),
            )
            for old_license in json_old.get("licenses")
        ]

        # filling the contributers section
        contributors = [
            structure.Contribution(
                contributor=structure.Person(
                    name=old_contributor.get("title"),
                    email=old_contributor.get("email"),
                ),
                date=parse_date(old_contributor.get("date")),
                obj=old_contributor.get("object"),
                comment=old_contributor.get("comment"),
            )
            for old_contributor in json_old.get("contributors")
        ]

        # extending with script-user information

        resources = []
        for resource in json_old.get("resources", []):
            fields = [
                structure.Field(
                    name=field.get("name"),
                    description=field.get("description"),
                    field_type=field.get("type"),
                    unit=field.get("unit"),
                )
                for field in resource["schema"].get("fields", [])
            ]
            field_dict = {field.name: field for field in fields}
            foreign_keys = []
            for fk in resource["schema"].get("foreignKeys"):
                source_fields = [
                    field_dict[field_name] for field_name in fk.get("fields", [])
                ]
                referenced_fields = [
                    structure.Field(
                        name=fk_field, unit=None, field_type=None, description=None
                    )
                    for fk_field in fk["reference"].get("fields")
                ]
                referenced_resource = structure.Resource(
                    name=fk["reference"].get("resource"),
                    schema=structure.Schema(
                        fields=referenced_fields, foreign_keys=None, primary_key=None
                    ),
                    dialect=None,
                    encoding=None,
                    path=None,
                    profile=None,
                    resource_format=None,
                )
                l = list()
                print(l)
                references = [
                    structure.Reference(s, t)
                    for s, t in zip(source_fields, referenced_fields)
                ]
                foreign_keys.append(structure.ForeignKey(references=references))
            schema = structure.Schema(
                fields=fields,
                primary_key=resource["schema"].get("primaryKey"),
                foreign_keys=foreign_keys,
            )

            dialect = structure.Dialect(
                delimiter=resource["dialect"].get("delimiter"),
                decimal_separator=resource["dialect"].get("decimalSeparator"),
            )
            resources.append(
                structure.Resource(
                    profile=resource.get("profile"),
                    name=resource.get("name"),
                    path=resource.get("path"),
                    resource_format=resource.get("format"),
                    encoding=resource.get("encoding"),
                    schema=schema,
                    dialect=dialect,
                )
            )

        inp_review = json_old["review"]
        review = structure.Review(
            path=inp_review.get("path"), badge=inp_review.get("badge")
        )

        inp_comment = json_old["_comment"]
        comment = structure.MetaComment(
            metadata_info=inp_comment.get("metadata"),
            dates=inp_comment.get("dates"),
            units=inp_comment.get("units"),
            languages=inp_comment.get("languages"),
            licenses=inp_comment.get("licenses"),
            review=inp_comment.get("review"),
            none=inp_comment.get("none"),
        )

        metadata = structure.OEPMetadata(
            name=json_old.get("name"),
            title=json_old.get("title"),
            identifier=json_old.get("id"),
            description=json_old.get("description"),
            languages=json_old.get("language"),
            keywords=json_old.get("keywords"),
            publication_date=parse_date(json_old.get("publicationDate")),
            context=context,
            spatial=spatial,
            temporal=temporal,
            sources=sources,
            terms_of_use=licenses,
            contributions=contributors,
            resources=resources,
            review=review,
            comment=comment,
        )
        return metadata

    def assert_1_3_metastring(self, json_string: str):
        """Checks string conformity to OEP Metadata Standard Version 1.3

        Parameters
        ----------
        json_string: str
            The JSON string to be checked.

        Returns
        -------
        bool
            True if valid, Raises Exception otherwise.
        """

        keys = [
            "title",
            "description",
            "language",
            "spatial",
            "temporal",
            "sources",
            "license",
            "contributions",
            "resources",
            "metadata_version",
        ]
        subkeys_spatial = ["location", "extent", "resolution"]
        subkeys_temporal = ["reference_date", "start", "end", "resolution"]
        subkeys_license = ["id", "name", "version", "url", "instruction", "copyright"]
        object_subkeys = {
            "spatial": subkeys_spatial,
            "temporal": subkeys_temporal,
            "license": subkeys_license,
        }
        subkeys_sources = [
            "name",
            "description",
            "url",
            "license",
            "copyright",
        ]  # in list of objects
        subkeys_contributors = [
            "name",
            "email",
            "date",
            "comment",
        ]  # in list of objects
        subkeys_resources = ["name", "format", "fields"]  # in list of objects
        list_subkeys = {
            "sources": subkeys_sources,
            "contributions": subkeys_contributors,
            "resources": subkeys_resources,
        }
        subkeys_resources_fields = ["name", "description", "unit"]  # in list of objects

        json_dict = json.loads(json_string)
        try:
            # check if all top level keys are present
            for i in keys:
                if not i in json_dict.keys():
                    raise Exception(
                        'The String did not contain the key "{0}"'.format(i)
                    )
            # check for all keys in second level objects
            for key in object_subkeys:
                for subkey in object_subkeys[key]:
                    if not subkey in json_dict[key]:
                        raise Exception(
                            'The "{0}" object did not contain a "{1}" key'.format(
                                key, subkey
                            )
                        )
            # check for all objects in lists if they contain all required keys
            for key in list_subkeys:
                for list_element in json_dict[key]:
                    for subkey in list_subkeys[key]:
                        if not subkey in list_element:
                            raise Exception(
                                'An object in "{0}" is missing a "{1}" key'.format(
                                    key, subkey
                                )
                            )
        except Exception as error:
            print(
                "The input String does not conform to metadatastring version 1.3 standard"
            )
            print(error)

    # TODO make function check all subkeys as well
    def has_rogue_keys(self, json_string):
        """Checks all keys if they are part of the metadata specification. Gives warnings if not.

        Parameters
        ----------
        json_string: str
            The JSON string to be checked.

        Returns
        -------
        """

        json_dict = json.loads(json_string)
        allowed_keys = [
            "title",
            "description",
            "language",
            "spatial",
            "temporal",
            "sources",
            "license",
            "contributions",
            "resources",
            "metadata_version",
        ]
        for j in json_dict.keys():
            if not j in allowed_keys:
                print('Warning: "{0}" is not among the allowed keys'.format(j))

    def get_table_name(self, metadata_file):
        """Provides the tablename information from the metadata_file

        Parameters
        ----------
        metadata_file: str
            The metadata script from where the tablename is extracted from.

        Returns
        -------
        tablename: str
            returns the tablename.
        """

        raise NotImplementedError



from omi.dialects.base.compiler import Compiler
from omi.dialects.base.parser import Parser
from omi.dialects.base.renderer import Renderer
from omi.structure import OEPMetadata


class Dialect:
    _parser = Parser
    _compiler = Compiler
    _renderer = Renderer

    def compile(self, obj: OEPMetadata, *args, **kwargs):
        c = self._compiler()
        return c.visit(obj, *args, **kwargs)

    def parse(self, string: str, *args, **kwargs) -> OEPMetadata:
        p = self._parser()
        return p.parse_from_string(string, *args, **kwargs)

    def render(self, structure, *args, **kwargs) -> str:
        r = self._renderer()
        return r.render(structure, *args, **kwargs)

    def compile_and_render(self, obj: OEPMetadata, *args, **kwargs):
        c = self._compiler()
        r = self._renderer()
        compiled = c.visit(obj, *args, **kwargs)
        return r.render(compiled, *args, **kwargs)



import pkgutil

from omi.dialects.base.register import get_dialect

for loader, module_name, is_pkg in pkgutil.walk_packages(__path__):
    n = "omi.dialects." + module_name + ".dialect"
    __import__(n)



from datetime import datetime
from enum import Enum
from typing import Iterable


class Compilable:
    __compiler_name__ = None

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__,
            ",".join("{}={}".format(key, val) for key, val in self.__dict__.items()),
        )

    def __lt__(self, other):
        for key in sorted(self.__dict__):
            s = getattr(self, key)
            o = getattr(other, key)
            if s is None:
                return True
            elif s < o:
                return True
            elif s > o:
                return False
        return False


class Language(Compilable):
    __compiler_name__ = "language"


class Spatial(Compilable):
    __compiler_name__ = "spatial"

    def __init__(self, location: str, extent: str, resolution: str):
        self.location = location
        self.extent = extent
        self.resolution = resolution


class TimestampOrientation(Compilable, Enum):
    __compiler_name__ = "timestamp_orientation"
    left = 0
    middle = 1
    right = 2

    @staticmethod
    def create(value: str) -> "TimestampOrientation":
        if value == "left":
            return TimestampOrientation.left
        elif value == "middle":
            return TimestampOrientation.middle
        elif value == "right":
            return TimestampOrientation.right
        else:
            raise Exception("Unknown timestamp orientation:", value)


class Temporal(Compilable):
    __compiler_name__ = "temporal"

    def __init__(
        self,
        reference_date: datetime,
        start: datetime,
        end: datetime,
        resolution: str,
        ts_orientation: TimestampOrientation,
    ):  # TODO: This should not be a string... maybe
        # we should use datetime instead?
        self.reference_date = reference_date
        self.ts_start = start
        self.ts_end = end
        self.ts_resolution = resolution
        self.ts_orientation = ts_orientation


class License(Compilable):
    __compiler_name__ = "license"

    def __init__(
        self,
        name: str,
        identifier: str,
        text: str,
        path: str,
        other_references: Iterable[str],
        comment: str = None,
    ):
        self.name = name
        self.path = path
        self.identifier = identifier
        self.other_references = other_references
        self.text = text
        self.comment = comment

    @staticmethod
    def instance_name_from_id(identifier: str):
        return "L_" + identifier.replace("-", "_").replace(".", "_").replace(
            "+", "_plus"
        )


class TermsOfUse(Compilable):
    __compiler_name__ = "terms_of_use"

    def __init__(self, instruction: str, attribution: str, lic: License):
        self.instruction = instruction
        self.attribution = attribution
        self.license = lic


class Source(Compilable):
    __compiler_name__ = "source"

    def __init__(
        self,
        title: str,
        description: str,
        path: str,
        source_license: License,
        source_copyright: str,
    ):
        self.title = title
        self.description = description
        self.path = path
        self.license = source_license
        self.copyright = source_copyright


class Person(Compilable):
    __compiler_name__ = "person"

    def __init__(self, name: str, email: str):
        self.name = name
        self.email = email


class Contribution(Compilable):
    __compiler_name__ = "contribution"

    def __init__(self, contributor: Person, date: datetime, obj: str, comment: str):
        self.contributor = contributor
        self.date = date
        self.object = obj
        self.comment = comment


class Field(Compilable):
    __compiler_name__ = "field"

    def __init__(
        self,
        name: str,
        description: str,
        field_type: str,
        unit: str,
        resource: "Resource" = None,
    ):
        self.name = name
        self.description = description
        self.type = field_type
        self.unit = unit
        self.resource = resource

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__,
            ",".join(
                "{}={}".format(key, val)
                for key, val in self.__dict__.items()
                if key != "resource"
            ),
        )


class Context(Compilable):
    __compiler_name__ = "context"

    def __init__(
        self,
        homepage: str,
        documentation: str,
        source_code: str,
        contact: str,
        grant_number: str,
    ):
        self.homepage = homepage
        self.documentation = documentation
        self.source_code = source_code
        self.contact = contact
        self.grant_number = grant_number


class Reference(Compilable):
    __compiler_name__ = "reference"

    def __init__(self, source: Field, target: Field):
        self.source = source
        self.target = target


class ForeignKey(Compilable):
    __compiler_name__ = "foreign_key"

    def __init__(self, references: Iterable[Reference]):
        self.references = references


class Schema(Compilable):
    __compiler_name__ = "schema"

    def __init__(
        self,
        fields: Iterable[Field],
        primary_key: Iterable[str],
        foreign_keys: Iterable[ForeignKey],
    ):
        self.fields = fields
        self.primary_key = primary_key
        self.foreign_keys = foreign_keys


class Dialect(Compilable):
    __compiler_name__ = "dialect"

    def __init__(self, delimiter: str, decimal_separator: str):
        self.delimiter = delimiter
        self.decimal_separator = decimal_separator


class Resource(Compilable):
    __compiler_name__ = "resource"

    def __init__(
        self,
        name: str,
        path: str,
        profile: str,
        resource_format: str,
        encoding: str,
        schema: Schema,
        dialect: Dialect,
    ):
        self.name = name
        self.path = path
        self.profile = profile
        self.format = resource_format
        self.encoding = encoding
        self.schema = schema
        if schema is not None:
            for field in schema.fields:
                field.resource = self
        self.dialect = dialect


class MetaComment(Compilable):
    __compiler_name__ = "meta_comment"

    def __init__(
        self,
        metadata_info: str,
        dates: str,
        units: str,
        languages: str,
        licenses: str,
        review: str,
        none: str,
    ):
        self.metadata_info = metadata_info
        self.dates = dates
        self.units = units
        self.languages = languages
        self.licenses = licenses
        self.review = review
        self.none = none


class Review(Compilable):
    __compiler_name__ = "review"

    def __init__(self, path: str, badge: str):
        self.path = path
        self.badge = badge


class OEPMetadata(Compilable):
    __compiler_name__ = "metadata"

    def __init__(
        self,
        name: str,
        title: str,
        identifier: str,
        description: str,
        languages: Iterable[Language],
        keywords: Iterable[str],
        publication_date: datetime,
        context: Context,
        spatial: Spatial,
        temporal: Temporal,
        sources: Iterable[Source],
        terms_of_use: Iterable[TermsOfUse],
        contributions: Iterable[Contribution],
        resources: Iterable[Resource],
        review: Review,
        comment: MetaComment,
    ):
        self.name = name
        self.title = title
        self.identifier = identifier
        self.description = description
        self.languages = languages
        self.keywords = keywords
        self.publication_date = publication_date
        self.context = context
        self.spatial = spatial
        self.temporal = temporal
        self.sources = sources
        self.license = terms_of_use
        self.contributions = contributions
        self.resources = resources
        self.review = review
        self.comment = comment

    def has_keywords(self):
        return self.keywords is not None



import json

from omi.dialects.oep.compiler import JSONCompiler
from omi.dialects.oep.parser import JSONParser_1_4
from omi.dialects.rdf.compiler import RDFCompiler
from omi.dialects.rdf.parser import RDFParser
from .test_compiler import assert_equal


def test_roundtrip():
    json_compiler = JSONCompiler()
    json_parser = JSONParser_1_4()
    rdf_compiler = RDFCompiler()
    rdf_p = RDFParser()
    with open("tests/data/metadata_v14.json", "r") as _input_file:
        input_string = _input_file.read()
        expected_json = json.loads(input_string)
        # Step 1: Parse JSON to internal structure
        internal_metadata = json_parser.parse_from_string(input_string)
        # Step 2: Translate to rdf
        _ = rdf_compiler.visit(internal_metadata)
        # Step 3: Parse rdf string
        internal_metadata2 = rdf_p.parse(rdf_compiler.graph)
        # Step 4: Translate to JSON
        result_json = json_compiler.visit(internal_metadata2)
        # Final step: Compare
        assert_equal(expected_json, result_json, disregard_ordering=True)



import json
from collections import OrderedDict

from omi import structure
from omi.dialects.base.compiler import Compiler


class JSONCompiler(Compiler):
    __METADATA_VERSION = "OEP-1.4"

    def visit_context(self, context: structure.Context, *args, **kwargs):
        return OrderedDict(
            homepage=self.visit(context.homepage),
            documentation=self.visit(context.documentation),
            sourceCode=self.visit(context.source_code),
            contact=self.visit(context.contact),
            grantNo=self.visit(context.grant_number),
        )

    def visit_contribution(self, contribution: structure.Contribution, *args, **kwargs):
        return OrderedDict(
            title=self.visit(contribution.contributor.name),
            email=self.visit(contribution.contributor.email),
            object=self.visit(contribution.object),
            date=contribution.date.strftime("%Y-%m-%d"),
            comment=self.visit(contribution.comment),
        )

    def visit_language(self, language: structure.Language, *args, **kwargs):
        return str(language)

    def visit_spatial(self, spatial: structure.Spatial, *args, **kwargs):
        return OrderedDict(
            location=self.visit(spatial.location),
            extent=self.visit(spatial.extent),
            resolution=self.visit(spatial.resolution),
        )

    def visit_timestamp_orientation(
        self, tso: structure.TimestampOrientation, *args, **kwargs
    ):
        if tso == structure.TimestampOrientation.left:
            return "left"
        elif tso == structure.TimestampOrientation.middle:
            return "middle"
        elif tso == structure.TimestampOrientation.right:
            return "right"
        else:
            raise NotImplementedError

    def visit_temporal(self, temporal: structure.Temporal, *args, **kwargs):
        start = None
        end = None
        if temporal.ts_start is not None:
            start = temporal.ts_start.strftime("%Y-%m-%dT%H:%M%z")[:-2]
        if temporal.ts_end is not None:
            end = temporal.ts_end.strftime("%Y-%m-%dT%H:%M%z")[:-2]
        return OrderedDict(
            referenceDate=temporal.reference_date.strftime("%Y-%m-%d"),
            start=start,
            end=end,
            resolution=self.visit(temporal.ts_resolution),
            timestamp=self.visit(temporal.ts_orientation),
        )

    def visit_source(self, source: structure.Source, *args, **kwargs):
        return OrderedDict(
            title=self.visit(source.title),
            description=self.visit(source.description),
            path=self.visit(source.path),
            license=self.visit(source.license.identifier) if source.license else None,
            copyright=self.visit(source.copyright),
        )

    def visit_license(self, lic: structure.License, *args, **kwargs):
        return OrderedDict(
            name=self.visit(lic.identifier),
            title=self.visit(lic.name),
            path=self.visit(lic.path),
        )

    def visit_terms_of_use(self, terms_of_use: structure.TermsOfUse):
        return OrderedDict(
            instruction=self.visit(terms_of_use.instruction),
            attribution=self.visit(terms_of_use.attribution),
            **self.visit(terms_of_use.license)
        )

    def visit_resource(self, resource: structure.Resource, *args, **kwargs):
        return OrderedDict(
            profile=self.visit(resource.profile),
            name=self.visit(resource.name),
            path=self.visit(resource.path),
            format=self.visit(resource.format),
            encoding=self.visit(resource.encoding),
            schema=self.visit(resource.schema),
            dialect=self.visit(resource.dialect),
        )

    def visit_field(self, field: structure.Field, *args, **kwargs):
        return OrderedDict(
            name=field.name,
            description=field.description,
            type=field.type,
            unit=field.unit,
        )

    def visit_schema(self, schema: structure.Schema, *args, **kwargs):
        return OrderedDict(
            fields=list(map(self.visit, schema.fields)),
            primaryKey=self.visit(schema.primary_key),
            foreignKeys=list(map(self.visit, schema.foreign_keys)),
        )

    def visit_dialect(self, dialect: structure.Dialect, *args, **kwargs):
        return OrderedDict(
            delimiter=self.visit(dialect.delimiter),
            decimalSeparator=self.visit(dialect.decimal_separator),
        )

    def visit_foreign_key(self, foreign_key: structure.ForeignKey, *args, **kwargs):
        if foreign_key.references:
            source_fields, target_fields, target_resources = zip(
                *map(self.visit, foreign_key.references)
            )

            target_resource = target_resources[0]

            return OrderedDict(
                fields=source_fields,
                reference=OrderedDict(
                    resource=self.visit(target_resource),
                    fields=self.visit(target_fields),
                ),
            )
        else:
            raise Exception("Missing reference in foreign key")

    def visit_reference(self, reference: structure.Reference, *args, **kwargs):
        return (
            reference.source.name,
            reference.target.name,
            reference.target.resource.name,
        )

    def visit_review(self, review: structure.Review, *args, **kwargs):
        return OrderedDict(path=review.path, badge=review.badge)

    def visit_meta_comment(self, comment: structure.MetaComment, *args, **kwargs):
        return OrderedDict(
            metadata=comment.metadata_info,
            dates=comment.dates,
            units=comment.units,
            languages=comment.languages,
            licenses=comment.licenses,
            review=comment.review,
            none=comment.none,
        )

    def visit_metadata(self, metadata: structure.OEPMetadata, *args, **kwargs):
        publication_date = None
        if metadata.publication_date is not None:
            publication_date = metadata.publication_date.strftime("%Y-%m-%d")
        return OrderedDict(
            name=metadata.name,
            title=metadata.title,
            id=metadata.identifier,
            description=metadata.description,
            language=list(map(self.visit, metadata.languages)),
            keywords=metadata.keywords,
            publicationDate=publication_date,
            context=self.visit(metadata.context),
            spatial=self.visit(metadata.spatial),
            temporal=self.visit(metadata.temporal),
            sources=list(map(self.visit, metadata.sources)),
            licenses=list(map(self.visit, metadata.license)),
            contributors=list(map(self.visit, metadata.contributions)),
            resources=list(map(self.visit, metadata.resources)),
            review=self.visit(metadata.review),
            metaMetadata=OrderedDict(
                metadataVersion=self.__METADATA_VERSION,
                metadataLicense=OrderedDict(
                    name="CC0-1.0",
                    title="Creative Commons Zero v1.0 Universal",
                    path="https://creativecommons.org/publicdomain/zero/1.0/",
                ),
            ),
            _comment=self.visit(metadata.comment),
        )
